/* -*- c++ -*- */
/*
 * Copyright 2016 The Aerospace Corporation.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "goes_sink_impl.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdexcept>
#define RE_ACQ

// win32 (mingw/msvc) specific
#ifdef _WIN32
//#ifdef HAVE_IO_H
#include <io.h>
//#endif
#endif // _WIN32

// for MSVC compiling
#ifdef _MSC_BUILD
#define _open_ _open
#define _close_  _close
#else
#define _open_ ::open
#define _close_ ::close
#endif // _MSC_BUILD

#ifdef O_BINARY
#define	OUR_O_BINARY O_BINARY
#else
#define	OUR_O_BINARY 0
#endif

// should be handled via configure
#ifdef O_LARGEFILE
#define	_O_LARGEFILE	O_LARGEFILE
#else
#define	_O_LARGEFILE 0
#endif

namespace gr {
  namespace goessdr {

    goes_sink::sptr
    goes_sink::make(int resynch_thresh, int delay_thresh,
                    const std::string &lritdir, const std::string &emwindir)
    {
      return gnuradio::get_initial_sptr(
                new goes_sink_impl(resynch_thresh, delay_thresh, lritdir, emwindir));
    }

    goes_sink_impl::goes_sink_impl(int resynch_thresh, int delay_thresh,
                                   const std::string &lritdir,
                                   const std::string &emwindir)
        : gr::sync_block("goes_sink",
			                 gr::io_signature::make(1, 1, sizeof(char)),
			                 gr::io_signature::make(0, 0, 0))
	    , d_resynch_thresh(resynch_thresh)
	    , d_delayed(false)  // delay should be initialized to 0
	    , d_rotated(false)
    {
      if (!d_lrit_processor.InitializeProcessor(
            boost::bind(&goes_sink_impl::toggle_viterbi_delay, this),
            boost::bind(&goes_sink_impl::toggle_rotate, this),
            delay_thresh))
	      throw std::runtime_error ("Can't initialize GoesProcessor");
      set_output_dirs(lritdir, emwindir);

      d_synched = true;

      // output port to control delay which toggles order of code bit 0
      // and code bit 1 into viterbi decoder
      message_port_register_out(pmt::mp("control"));
      message_port_register_out(pmt::mp("report"));
      message_port_register_in(pmt::mp("reset"));
      message_port_register_in(pmt::mp("report"));
      set_msg_handler(pmt::mp("reset"), boost::bind(&goes_sink::reset_error_counts, this, _1));
      set_msg_handler(pmt::mp("report"), boost::bind(&goes_sink::report_error_counts, this, _1));
    }

    goes_sink_impl::~goes_sink_impl()
    { }

    bool goes_sink_impl::get_synch()
    {
      return d_synched;
    }

    int
    goes_sink_impl::work (int noutput_items,
			   gr_vector_const_void_star &input_items,
			   gr_vector_void_star &output_items)
    {
      unsigned char *inbuf = (unsigned char *) input_items[0];
      int  nwritten = 0;

#ifdef RE_ACQ
      //d_synched=true;
      d_synched = d_lrit_processor.ProcessLritDataBlock(inbuf, noutput_items, d_resynch_thresh);
      if(d_synched) {
	      if (noutput_items==0) {
	        printf( "noutput_items = 0 \n");
	      }
	    return noutput_items;
      }
      else {
      	printf( "LRIT Image_sink PROCESSOR:: Not synched \n");
      	return -1;
      }
#else
      d_lrit_processor.ProcessLritDataBlock(inbuf, noutput_items, d_resynch_thresh);
      if (noutput_items==0) {
      	printf( "noutput_items = 0 \n");
      }
      return noutput_items;
#endif
    }

    void goes_sink_impl::toggle_viterbi_delay()
    {
        GR_LOG_DEBUG(d_debug_logger,
                    boost::format("Viterbi Delay: %d --> %d") %
                                  d_delayed % !d_delayed);

        d_delayed = !d_delayed;
        pmt::pmt_t msg = pmt::make_dict();
        pmt::pmt_t key = pmt::mp("dly");
        pmt::pmt_t value = pmt::make_tuple(pmt::mp((long)d_delayed));
        msg = pmt::dict_add(msg, key, value);
        GR_LOG_DEBUG(d_debug_logger,
                    boost::format("Send message: %s") % pmt::write_string(msg));
        message_port_pub(pmt::mp("control"), msg);
    }

    void goes_sink_impl::toggle_rotate()
    {
        GR_LOG_DEBUG(d_debug_logger,
                    boost::format("ROTATING %d DEGREES") %
                    (d_rotated? 90: 0))

        d_rotated != d_rotated;
        pmt::pmt_t msg = pmt::make_dict();
        pmt::pmt_t key = pmt::mp("k");
        pmt::pmt_t value = pmt::make_tuple(pmt::make_rectangular(!d_rotated, d_rotated));
        msg = pmt::dict_add(msg, key, value);
        GR_LOG_DEBUG(d_debug_logger,
                    boost::format("Send message: %s") % pmt::write_string(msg));
        message_port_pub(pmt::mp("control"), msg);
    }

    void goes_sink_impl::reset_error_counts(const pmt::pmt_t &msg)
    {
        if (pmt::is_true(msg)) d_lrit_processor.ResetErrorCounts();
    }

    void goes_sink_impl::report_error_counts(const pmt::pmt_t &msg)
    {
        if (pmt::is_true(msg)) {
          float total_fer, rs_fer, vit_ber;
          int vit_err, rs_err, vcdu_miss, tot_vcdu, lock_level;
          // all values passed by reference and filled in
          d_lrit_processor.ReportErrorCounts(total_fer, rs_fer, vit_ber,
                                           vit_err, rs_err, vcdu_miss, tot_vcdu,
                                           lock_level);
          /*
          // using a dict allows labels to go along with value
          pmt::pmt_t d = pmt::make_dict();
          pmt::dict_add(d, pmt::mp("TOT FER"),    pmt::mp(total_fer));
          pmt::dict_add(d, pmt::mp("RS FER"),     pmt::mp(rs_fer));
          pmt::dict_add(d, pmt::mp("VIT BER"),    pmt::mp(vit_ber));
          pmt::dict_add(d, pmt::mp("MISS VCDU"),  pmt::mp(vcdu_miss));
          pmt::dict_add(d, pmt::mp("TOT VCDU"),   pmt::mp(tot_vcdu));
          pmt::dict_add(d, pmt::mp("LOCK STATE"), pmt::mp(lock_level));
          */
          // using a tuple ensures a fixed ordering
          pmt::pmt_t tup = pmt::make_tuple(pmt::mp(total_fer),
                                           pmt::mp(rs_fer),
                                           pmt::mp(vit_ber),
                                           pmt::mp(vcdu_miss),
                                           pmt::mp(tot_vcdu),
                                           pmt::mp(lock_level));
          pmt::pmt_t msg = pmt::make_dict();
          msg = pmt::dict_add(msg, pmt::mp("goes errors"), tup);
          message_port_pub(pmt::mp("report"), msg);
        }
    }

    void goes_sink_impl::set_output_dirs(const std::string &ldir, const std::string &edir)
    {
        d_lrit_processor.SetOutputDirs(ldir, edir);
    }

  } /* namespace dcid */
} /* namespace gr */
