"""
@brief CPPDU (Transport Frame) Message class.
@details Contains a single Transport Frame. The most useful
header items are packet length and apid (for constructing the transport
frame) and sequence flag (for using the frame to construct a file)
@author Miguel Dominguez
@date 2016
"""

import logging
from GOESFrame import GOESFrame

class CPPDUMsg(GOESFrame):

    VERSION_NUM_MASK = 0xE0
    VERSION_NUM_SHIFT = 5

    TYPE_MASK = 0x10
    TYPE_SHIFT = 4

    SECONDARY_HEADER_FLAG = 0x08
    SECONDARY_HEADER_SHIFT = 3

    APID_MASK_LEFT = 0x07
    APID_MASK_SHIFT_LEFT = 8
    APID_MASK_RIGHT = 0xFF

    SEQUENCE_FLAG_MASK = 0xC0
    SEQUENCE_FLAG_SHIFT = 6

    PACKET_SEQUENCE_COUNT_MASK_LEFT = 0x3F
    PACKET_SEQUENCE_COUNT_SHIFT_LEFT = 8
    PACKET_SEQUENCE_COUNT_MASK_RIGHT = 0xFF

    PACKET_LENGTH_MASK = 0xFF
    PACKET_LENGTH_SHIFT_LEFT = 8

    CRC_MASK = 0xFF
    CRC_SHIFT = 8

    HEADER_SIZE_BYTES = 6
    CRC_SIZE_BYTES = 2

    APID_FILL_PACKET =  2047

    MAX_PACKET_LENGTH = 8192

    ##
    # @fn __init__
    # @brief Constructor.
    # @details Initializes the logger and an empty data list
    def __init__(self):
        self.data = []
        self.crc = None
        self.logger = logging.getLogger(__name__)

    ##
    # @fn getHeader
    # @brief Extracts the header values with appropriate bit masking and shifting
    # @param[in] headerBytes Unsigned byte list sufficiently long enough to
    # contain the header (6 bytes)
    def getHeader(self,headerBytes):
#        print(' '.join('{:02x}'.format(x) for x in headerBytes[0:6]))
        self.versionNum = (headerBytes[0] & self.VERSION_NUM_MASK) >> self.VERSION_NUM_SHIFT
        self.type = (headerBytes[0] & self.TYPE_MASK) >> self.TYPE_SHIFT
        self.secondaryHeaderFlag = (headerBytes[0] & self.SECONDARY_HEADER_FLAG) >> self.SECONDARY_HEADER_SHIFT
        self.apid = (headerBytes[0] & self.APID_MASK_LEFT) << self.APID_MASK_SHIFT_LEFT
        self.apid |= (headerBytes[1] & self.APID_MASK_RIGHT)
        self.sequenceFlag = (headerBytes[2] & self.SEQUENCE_FLAG_MASK) >> self.SEQUENCE_FLAG_SHIFT
        self.packetSequenceCount = (headerBytes[2] & self.PACKET_SEQUENCE_COUNT_MASK_LEFT) << self.PACKET_SEQUENCE_COUNT_SHIFT_LEFT
        self.packetSequenceCount |= (headerBytes[3] & self.PACKET_SEQUENCE_COUNT_MASK_RIGHT)
        self.packetLength = (headerBytes[4] & self.PACKET_LENGTH_MASK) << self.PACKET_LENGTH_SHIFT_LEFT
        self.packetLength |= (headerBytes[5] & self.PACKET_LENGTH_MASK)
        #For some reason packetLength is subtracted by 1 when encoded by TX
        self.packetLength += 1

    ##
    # @fn fromFrame
    # @brief Extracts the MPDU Message from the VCDUMsg
    # @details Unlike the other messages, does NOT call getHeader automatically
    # @param[in] data VCDUMsg
    def fromFrame(self,data):
        self.data += data

    ##
    # @fn getCrcFromData
    # @brief Extracts the CRC from the internal data varaible
    # @warning This assumes the last two bytes are the CRC, which is a bad
    # assumption if the frame is incomplete when this function is called!
    def getCrcFromData(self):
        #Cut CRC out of data and put it into its own field
        if len(self.data) < 2:
            self.logger.error('No data found in Transport Frame!')
            #Should never match any valid CRC (which are unsigned)
            self.crc = -1
        else:
            crcbytes = self.data[-self.CRC_SIZE_BYTES:]
            self.data = self.data[:-self.CRC_SIZE_BYTES]
            #Make CRC a single word for convenience
            self.crc = (crcbytes[0] & self.CRC_MASK) << self.CRC_SHIFT
            self.crc |= (crcbytes[1] & self.CRC_MASK)

    ##
    # @fn __str__
    # @brief Returns formatted header information as a string
    # @param[out] str Header information as a string
    def __str__(self):
        output = "Version Number: {0}\n".format(self.versionNum)
        output += "Type: {0}\n".format(self.type)
        output += "Secondary Header Flag: {0}\n".format(self.secondaryHeaderFlag)
        output += "APID: {0}\n".format(self.apid)
        output += "Sequence Flag: {0}\n".format(self.sequenceFlag)
        output += "Packet Sequence Count: {0}\n".format(self.packetSequenceCount)
        output += "Packet Length: {0}\n".format(self.packetLength)
 #      output += "CRC: {0}\n".format(self.crc)
 #      output += "Actual Packet Length: {0}\n".format(len(self.data) + 2)
     #  output += "".join([chr(x) for x in self.data])
        return output
