#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

from gnuradio import gr, gr_unittest
from gnuradio import blocks
from message_passing_control_wrapper import message_passing_control_wrapper
import numpy as np
import pmt
import time
import sys

class message_poster(gr.sync_block):
  def __init__(self):
    super(message_poster, self).__init__(
      "message_poster", None, None)
    self.mport = mport = "pub_port"
    self.message_port_register_out(pmt.intern(mport))

  def post_message(self, msg):
    self.message_port_pub(pmt.intern(self.mport), msg)

class qa_message_passing_control_wrapper (gr_unittest.TestCase):

    def setUp (self):
        self.tb = gr.top_block ()

    def tearDown (self):
        self.tb = None

    def test_print(self):
        """test print port for debug"""
        itemsize = gr.sizeof_gr_complex
        N = 100
        sps = 1
        thr = message_passing_control_wrapper(blocks.throttle, itemsize, sps)
        src = blocks.vector_source_c(np.exp(2j * np.pi * np.arange(N) / float(N)), repeat=True)
        snk = blocks.null_sink(itemsize)
        self.tb.connect(src, thr, snk)
        poster = message_poster()
        # only proxy_blk exposes print_port
        self.tb.msg_connect(poster, poster.mport, thr.proxy_blk, thr.proxy_blk.print_port)
        self.tb.start()
        for n in range(10):
          time.sleep(0.1)
#          thr.proxy_blk.to_basic_block()._post(pmt.intern(thr.proxy_blk.ping_port),
#                                               pmt.to_pmt({'One': 'MESSAGE', 'Two': n}))
          poster.post_message(pmt.to_pmt('Message %d' % n))
          
        time.sleep(1)
        self.tb.stop()
        self.tb.wait()
        self.assertTrue(hasattr(sys.stdout, 'getvalue'), "Must run test with buffer=True.")
        lines = sys.stdout.getvalue()
        self.assertEqual(''.join('RECEIVED: Message %d\n' % n for n in range(10)), lines) 

    def test_ports(self):
        """test message port counts of hier and proxy blocks"""
        itemsize = gr.sizeof_gr_complex
        dly = message_passing_control_wrapper(blocks.delay, itemsize, 0)
        # hier block has no explicit message ports (see qa_hier_block2_message_connections.py)
        self.assertEqual(pmt.length(dly.message_ports_in()), 0)
        # only proxy block has countable ports -- 3 connections: 
        # "system" is for system level control
        # .set_port is for setting parameters
        # .print_port is for debugging messages
        self.assertEqual(pmt.length(dly.proxy_blk.message_ports_in()), 3)
        
    def test_passthrough (self):
        """test pass through for wrapped delay"""
        # set up fg: 
        # sine -> throttle -> delay -> mul -> sink
        #               \              /
        #                ---> conj ----
        itemsize = gr.sizeof_gr_complex
        M, N = 10, 100   # num blocks, block size
        sps = M * N      # sample rate (1s run time)
        src = blocks.vector_source_c(np.exp(2j * np.pi * np.arange(M * N) / float(N)), repeat=False)
        snk = blocks.vector_sink_c()
        cnj = blocks.conjugate_cc()
        thr = blocks.throttle(itemsize, sps)
        mul = blocks.multiply_cc()
        dly = message_passing_control_wrapper(blocks.delay, itemsize, 0)
        self.tb.connect(src, thr, dly, (mul, 0))
        self.tb.connect(src, cnj, (mul, 1))
        self.tb.connect(mul, snk)
        self.tb.run ()
        # check data
        data = snk.data()
        self.assertTrue(np.allclose(data, [1.0,] * (M * N)))

    def test_shifted (self):
        """test run-time change of delay"""
        # set up fg: 
        # sine -> throttle -> delay -> sub -> sink
        #               \              /
        #                --------------
        itemsize = gr.sizeof_gr_complex
        M, N = 10, 20   # num blocks, block size
        sps = 100 * M * N      # sample rate (0.01s per frame)
        src = blocks.vector_source_c(np.exp(2j * np.pi * np.arange(M * N) / float(N)), repeat=True)
        stm2vec = blocks.stream_to_vector(itemsize, M * N)
        snk = blocks.probe_signal_vc(M * N)
        cnj = blocks.conjugate_cc()
        thr = blocks.throttle(itemsize, sps)
        mul = blocks.multiply_cc()
        dly = message_passing_control_wrapper(blocks.delay, itemsize, 0)
        self.tb.connect(src, thr, dly, (mul, 0))
        self.tb.connect(src, cnj, (mul, 1))
        self.tb.connect(mul, stm2vec, snk)
        poster = message_poster()
        self.tb.msg_connect(poster, poster.mport, dly, dly.set_port)
        delay = 4
        self.tb.start()
        # TODO: post this way or with message strobe?
        # must post to actual block, message strobe can connect to hier
        #dly.proxy_blk.to_basic_block()._post(pmt.intern(dly.proxy_blk.ping_port), 
        #                                     #pmt.to_pmt({'dly': delay})
        #                                     pmt.to_pmt("STRING"))
        time.sleep(0.1)
        predelay = snk.level()
        time.sleep(0.1)
        poster.post_message(pmt.to_pmt({'dly': [delay,]}))
        while dly.blk.dly() != delay:
          time.sleep(0.1)
          #print 'delay = %d' % dly.blk.dly()
        postdelay = snk.level()
        self.tb.stop()
        # proper way of telling a pure message passing block to stop
        # https://lists.gnu.org/archive/html/discuss-gnuradio/2015-05/msg00288.html
        #dly.proxy_blk.to_basic_block()._post(pmt.intern('system'),
        #                                     pmt.to_pmt({'done': 1L}))
        # TODO: pmt.to_pmt({x:y}) gives something different than
        # pmt.cons(pmt.to_pmt(x), pmt.to_pmt(y)) which equals
        # pmt.cars(pmt.to_pmt({x:y})).  Which one works as intended?        
        self.tb.wait()
        # check data
        self.assertComplexTuplesAlmostEqual((1.0 + 0j, ) * len(predelay), predelay)
        self.assertComplexTuplesAlmostEqual(
          (np.exp(-2j * np.pi * delay / float(N)),) * len(postdelay), postdelay)


if __name__ == '__main__':
    #gr_unittest.run(qa_message_passing_control_wrapper 
    #               ,"qa_message_passing_control_wrapper.xml"
    #               )
    gr_unittest.main(buffer=True, verbosity=0)
