#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

import sys
import os
import os.path as osp

import logging
logging.basicConfig(level=logging.WARN,
                    format='%(asctime)s - %(levelname)s - ' #%(module)s.%(funcName)s -
                           '%(message)s')
                    #datefmt='%m/%d/%Y %H:%M:%S.%f')    # comment out to use ISO8601 time format
logger = logging.getLogger(osp.splitext(osp.basename(__file__))[0])

import threading as th
import numpy as np

import time
import pmt

from gnuradio import gr
gr.logger_config(osp.expanduser(osp.join('~', '.gnuradio', 'gr_log_default.prop')))

from warnings import warn

K, M, G = tuple(10. ** (3 * x) for x in range(1, 4))

# this is useful for older GR which don't provide the argument_parser function (deprecated?)
def get_static_argparser():
    parser = OptionParser(option_class=eng_option, usage="%prog: [options]")
    parser.add_option("", "--decim", dest="decim", type="intx", default=1,
        help="Set Channelizer decimation factor [default=%default]")
    parser.add_option(
        "", "--fc", dest="fc", type="eng_float", default=eng_notation.num_to_137(5e3),
        help="Set Carrier frequency [default=%default]")
    parser.add_option(
        "", "--pll-loopbw", dest="pll_loopbw", type="eng_float", default=eng_notation.num_to_str(0.02),
        help="Set Freq detect loop bw (norm'd) [default=%default]")
    parser.add_option(
        "", "--rrc-order", dest="rrc_order", type="intx", default=64,
        help="Set filter order [default=%default]")
    parser.add_option(
        "", "--samp-rate", dest="samp_rate", type="eng_float", default=eng_notation.num_to_str(2e6),
        help="Set sample rate [default=%default]")
    parser.add_option(
        "", "--symb-rate", dest="symb_rate", type="eng_float", default=eng_notation.num_to_str(500e3),
        help="Set symbol rate [default=%default]")
    return parser
    
def add_parser_args(parser):
    parser.add_option("", "--stats-interval", type="float", default=5,
                      help="Seconds between statistics messages [0 = off]")
    parser.add_option("-v", "--verbose", action='count', help="Increase log verbosity [default=WARN]")
    parser.add_option("-q", "--quiet", action='count', help="Decrease log verbosity [default=WARN]")
    return parser

# the following setup was taken from the GRC generated Qt flowgraph 
qapp = None
def get_tb(tbfg, *args, **kwds):
    global qapp
    
    from PyQt4 import Qt
    if Qt.QWidget in tbfg.__bases__:
        # taken from GRC generated output
        from distutils.version import StrictVersion
        if StrictVersion(Qt.qVersion()) >= StrictVersion("4.5.0"):
            style = gr.prefs().get_string('qtgui', 'style', 'raster')
            Qt.QApplication.setGraphicsSystem(style)
        qapp = Qt.QApplication(sys.argv)
        
    tb = tbfg(*args, **kwds)
    
    try:
        def quitting():
            tb.stop()
            tb.wait()
        qapp.connect(qapp, Qt.SIGNAL("aboutToQuit()"), quitting)
    except AttributeError:  # not a Qt flowgraph
        pass
    return tb

def main(*args):
    global logger
    
    import goes_lrit_emwin_fg as goesfg
    
    # add some arguments to flowgraph arg parser
    try:
        argparser = goesfg.argument_parser()
    except AttributeError:
        logger.warn('Out-dated GR does not provide argument_parser() function,'
                    ' receiver options may be out-of-date.')
        argparser = get_static_argparser()
        
    argparser = add_parser_args(argparser)
    options, posargs = argparser.parse_args(args=list(args))

    logger.setLevel(logger.getEffectiveLevel() - 
                    10 * ((options.verbose or 0) - (options.quiet or 0)))
    logger.critical("Log level = %s" % logging.getLevelName(logger.getEffectiveLevel()))

    # flowgraph arguments don't include local flags (better way to do this?)
    fgargs = {k: v for k, v in options.__dict__.iteritems() if k not in ['stats_interval', 'verbose', 'quiet']}

    # create top block
    tb = get_tb(goesfg.goes_lrit_emwin_fg, **fgargs)

    # output some info
    logger.critical("Receiver settings:")
    logger.critical("Initialy carrier = %.3f MHz" % (tb.get_fc() / M))
    logger.critical("Sample rate = %.3f Msps" % (tb.get_samp_rate() / M))
    logger.critical("Symbol rate = %.3f Kbaud" % (tb.get_symb_rate() / K))
    logger.critical("Digital decimation = %d" % tb.get_decim())
    logger.critical("Timing samples-per-symbol = %.2f" % tb.get_sps())

    set_fo = lambda f: tb.fir_channelizer.to_basic_block()._post(pmt.intern('freq'),
                                                                 pmt.to_pmt(f))

    fo_control = FoffsetControl(logger, tb.get_fc(), tb.get_samp_rate(), tb.get_sps(),
                                set_fo, tb.osmosdr_source_0.set_center_freq if
                                        hasattr(tb, 'osmosdr_source_0') else None)

    # thread to monitor frequency offset
    running = th.Event()
    th_foffset = ThreadCorrectFoffset(running, logger, tb.get_samp_rate(),
                                      fo_control.set_foff, tb.blocks_probe_signal_x_0.level,
                                      interval=1,
                                      off_tol=0.005 * tb.get_symb_rate(),
                                      on_tol=0.02 * tb.get_symb_rate())
    th_foffset.start()

    th_snr = ThreadEstimateSNR(running, logger, 
                               tb.digital_probe_mpsk_snr_est_c_0.snr,
                               interval=options.stats_interval)
    th_snr.start()

    th_datarate = ThreadEstimateRate(running, logger, 
                                     tb.blocks_probe_rate_0.rate,
                                     interval=options.stats_interval)
    th_datarate.start()

    goes_sink_bb = tb.goessdr_goes_sink_0.to_basic_block()
    th_err = ThreadReportErrors(running, logger, 
                                lambda: goes_sink_bb._post(pmt.to_pmt('report'), pmt.PMT_T),
                                interval=options.stats_interval)
    th_err.start()
    
    try:
        tb.start()
        running.set()
        try:
            tb.show()
            qapp.exec_()
        except AttributeError:
            # not a Qt flowgraph
            # TODO: do we want this only w/o qt GUI, put into stdin thread?
            while sys.stdin:
                cmd = raw_input()
                if not cmd or cmd.lower() in 'q': break
                if cmd.lower() in 'h?':
                    print CMD_HELP
                elif cmd.lower() in 'r':
                    goes_sink_bb._post(pmt.to_pmt('reset'), pmt.PMT_T)
                elif cmd.lower() in 'p':
                    goes_sink_bb._post(pmt.to_pmt('report'), pmt.PMT_T)
            
        # TODO: This code requires hitting enter to send the char; try
        # http://stackoverflow.com/questions/510357/python-read-a-single-character-from-the-user
        # for operating on a single character with no enter
    except EOFError, KeyboardInterrupt:
        pass
    finally:
        logger.critical("Exiting...")
        running.clear()
        try:
            qapp.quit()
        except AttributeError:
            tb.stop()
            tb.wait()
        th_foffset.join()
        th_snr.join()
        th_datarate.join()
        th_err.join()
        sys.stdout.flush()
        sys.stderr.flush()
        
    return 0

CMD_HELP = """
####### USAGE #######################################
?, h        Print this help message
r           Reset error counters
p           Print error counters
ctrl-d, q   Stop the receiver and exit
####### USAGE #######################################
"""

class FoffsetControl(object):
    """
    Set frequency offset correction via combination of front-end tuning and
    digital channel tuning.
    """
    def __init__(self, logger, fc, fs, fx, fcn_fo, fcn_fc, guard_band=0.1):
        self.fc = fc                        # Hz, initial carrier freq
        self.fs = fs                        # Hz, sample rate
        self.fx = fx                        # oversampling rate
        self.band_edge = 0.5 - guard_band   # (normed) band edge before retuning hardware
        self.foff = 0
        self.fcn_set_fo = fcn_fo
        self.fcn_set_fc = fcn_fc
        self.logger = logger
        
    # foff is in Hz
    def set_foff(self, foff, alpha=1.):
        over_band_edge = abs(self.foff + foff) / self.fs + 0.5 / self.fx > self.band_edge
        #print abs(self.foff + foff) / self.fs + 0.5 / self.fx, '<>' , self.band_edge
        if self.fcn_set_fc and over_band_edge:
            self.fc += self.foff + foff    # recenter carrier in hardware
            self.logger.info("Set carrier = %.3f KHz" % (self.fc / K))
            self.fcn_set_fc(self.fc)
            self.foff = 0
        elif not over_band_edge:
            self.foff += alpha * foff   # TODO: should we have an alpha factor < 1 here?
        else:
            self.foff = -1 ** (foff < 0) * (self.band_edge - 0.5 * self.fx) * self.fs
            
        self.logger.info("Set digital offset = %.3f KHz", self.foff / K)
        self.fcn_set_fo(self.foff)

# TODO: should this be limited to 1 round of corrections at the start of the run?
# Perhaps tracking should be done digitally in flowgraph?  If only probe had a message out
class ThreadCorrectFoffset(th.Thread):
    """Monitor and corrects frequency offset.
    
       Uses fcn_set_fc to change the carrier freq and fcn_probe_fo to detect
       the measured fc.  fc is the initial set carrier frequency.
    """
    def __init__(self, running, logger, fs, fcn_set_foff, fcn_meas_foff,
                 interval=1, go_tol=0.001, off_tol=500, on_tol=5000):
        super(ThreadCorrectFoffset, self).__init__(name='goes_thread_foffset')
        self.fs = fs                        # Hz, sample rate
        self.fcn_set_foff = fcn_set_foff    # Hz, new freq offset
        self.fcn_meas_foff = fcn_meas_foff      # measure center frequency function
        self.interval = interval
        # the following offer a hysteresis, of sorts, to the freq tracking process
        self.off_tol = self.tol = off_tol   # tolerance to stop adjusting freq
        self.on_tol = on_tol                # tolerance to start adjusting freq
        self.isoff = th.Event()             # False initially

        self.running = running
        self.logger = logger

    def run(self):
        self.running.wait()
        while self.running.is_set():
            time.sleep(self.interval)
            foff = self.fcn_meas_foff()     # measurement returns foffset (Hz)
            self.logger.debug("Measured carrier offset = %.3f KHz", foff / K)
            if abs(foff) > self.tol:
                if self.isoff.is_set():
                    self.tol = self.off_tol
                    self.isoff.clear()
            else:
                self.tol = self.on_tol
                self.isoff.set()
            if not self.isoff.is_set():
                self.fcn_set_foff(foff)

class ThreadEstimateSNR(th.Thread):
    """Monitor SNR estimates.
    
       Uses fcn_meas_snr returns a linear SNR estimate.
    """
    def __init__(self, running, logger, fcn_meas_snr, interval=5):
        super(ThreadEstimateSNR, self).__init__(name='goes_thread_estsnr')
        self.running = running
        self.logger = logger
        self.fcn_meas_snr = fcn_meas_snr
        self.interval = interval
        
    def run(self):
        if not self.interval: return    # do not run if explicitly turned off
        self.running.wait()
        while self.running.is_set() and self.interval:
            time.sleep(self.interval)
            # TODO: current M2M4 MPSK SNR estimator is for complex signals;
            # for real signals it differs, however the error rate agrees with 
            # the measured SNR as is
            self.logger.info("Measured snr = %.3f dB", self.fcn_meas_snr())

class ThreadEstimateRate(th.Thread):
    """Monitor SNR estimates.
    
       Uses fcn_meas_rate returns a bps estimate (post FEC).
    """
    def __init__(self, running, logger, fcn_meas_rate, interval=5):
        super(ThreadEstimateRate, self).__init__(name='goes_thread_estrate')
        self.running = running
        self.logger = logger
        self.fcn_meas_rate = fcn_meas_rate
        self.interval = interval

    def run(self):
        if not self.interval: return    # do not run if explicitly turned off
        self.running.wait() 
        while self.running.is_set():
            time.sleep(self.interval)
            self.logger.info("Uncoded data rate = %.1f Kbps", self.fcn_meas_rate() / K)

class ThreadReportErrors(th.Thread):
    """Monitor errors and error rates.
    
       Uses fcn_trigger_report triggers a log report of the errors.
    """
    def __init__(self, running, logger, fcn_trigger_report, interval=5):
        super(ThreadReportErrors, self).__init__(name='goes_thread_report_err')
        self.running = running
        self.fcn_trigger_report = fcn_trigger_report
        self.interval = interval
        
    def run(self):
        if not self.interval: return    # do not run if explicitly turned off
        self.running.wait()
        while self.running.is_set() and self.interval:
            time.sleep(self.interval)
            self.fcn_trigger_report()
        
if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))

