#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

# requires Qt for some functionality
from traits.trait_base import ETSConfig
ETSConfig.toolkit = 'qt4'

from traits.api import *
from traitsui.api import *

from pyface.qt.QtGui import QDesktopServices as qds
from pyface.qt.QtCore import QUrl, QDir, QFileSystemWatcher

from threading import Thread, Lock
from os import listdir, getcwd
import os.path as osp
from qfswatcher import QFSWatcher

from traitsui.file_dialog import FileInfo
from traitsui.tabular_adapter import TabularAdapter

import basic_configurations_tab as bct
import config

# FileInfo is an "extension" used to show additional info in an open file dialog;
# Additional ones may be defined (TODO: preview extension showing thumbnails?)
class FileEntry(FileInfo):
    """Customized FileInfo object"""
    basename = Property(depends_on='file_name')
    file_size = Property(depends_on='file_name')

    traits_view = View(Item('basename', label='Name'),
                      Item('file_size', label='Size'),
                      Item('ctime', label='Creation'),
                      style='readonly'
                      )


    def _get_basename(self):
        try: return osp.basename(self.file_name)
        except: pass
        return ''
        
    def _get_file_size(self): 
        try: return osp.getsize(self.file_name)
        except: pass
        return 0

class FileListAdapter(TabularAdapter):
    columns = [('Name', 'basename'),
               ('Size', 'file_size'),
               ('Date', 'ctime')]
    alignment = 'left'

class ScanReadFiles(HasTraits):
    label=Str
    # Directory trait returns unicode string which is incompatible with std::string
    # used in c++ objects (e.g. the flowgraph) so use CStr
    directory = CStr(label='Directory', editor=DirectoryEditor(invalid='directory_invalid'),
                     section="Scan and Read Files", config_name="Directory")
    directory_invalid = Property(Bool, depends_on='directory')
    dclicked = Event()
    
    file_list = List(FileEntry)
    file_counter = Property(Long, label='File count')

    filters = Instance(bct.BasicConfigurationsTab)
    filtered_list = Property(List, editor=TabularEditor(
                                        adapter=FileListAdapter(),
                                        editable=False,
                                        operations=[],
                                        vertical_lines=False,
                                        auto_resize=True,
                                        dclicked='dclicked'))
    filtered_counter = Property(Long, label='Filtered count')

    def _directory_default(self):
        return config.HOME_DIRECTORY

    @property_depends_on('file_list')
    def _get_file_counter(self): 
        return len(self.file_list)

    @property_depends_on('filtered_list')
    def _get_filtered_counter(self):
        return len(self.filtered_list)

    def _dclicked_fired(self, entry):
        #import ipdb_for_qt; ipdb_for_qt.set_trace()
        # entry is a TabularEditorEvent in traitsui.traitsui.qt4.editors.tabular_editor
        if entry is not None:
            qds.openUrl(QUrl('file://' + entry.item.file_name))

    def get_file_list(self):
        fullpaths = (osp.join(self.directory, fn) for fn in listdir(self.directory))
        return [FileEntry(file_name=fn) for fn in fullpaths if osp.isfile(fn)]
    
    def _file_list_default(self): return self.get_file_list()

    def update_filelist(self, new):
        # currently we have 1 handler watching all directories and notifying 
        # everyone when any change, so update only for this directory
        if new == self.directory and not self.directory_invalid:
            self.file_list = self.get_file_list()

    def _get_directory_invalid(self): return not osp.isdir(self.directory)

    # EMWIN filenames are PPPOOOSS.EXT where PPP = product abbrev, OOO = office 
    # abbrev, SS = state abbrev, and EXT is some filetype-based extension (can this be assumed 3 chars?)
    @property_depends_on('file_list, filters.[state_filter.selected, counties_filter, product_filter, office_filter]')
    def _get_filtered_list(self):
        myfilters = self.filters
              
        # setup a filter pipeline: mark all emwin files, keep all non-emwin files
        # TODO: is this sufficient to get all EMWIN files?
        # TODO: naming convention changes for HRIT/EMWIN, check http://www.nws.noaa.gov/emwin/index.htm
        filenames = (fentry.basename for fentry in self.file_list)
        markemwins = ((fn, len(fn) == 12 and fn[-4] == '.') for fn in filenames)
        # *** NOTE: if filter is empty, assume ALL, i.e. no filter ***
        if myfilters.product_abbrev:
            byproduct = ((fn, isemwin) for fn, isemwin in markemwins 
                            if not isemwin or fn[:3] in myfilters.product_abbrev)
        else:
            byproduct = markemwins
            
        if myfilters.office_abbrev:
            byoffice = ((fn, isemwin) for fn, isemwin in byproduct 
                            if not isemwin or fn[3:6] in myfilters.office_abbrev)
        else:
            byoffice = byproduct
            
        if myfilters.state_abbrev:
            bystate = ((fn, isemwin) for fn, isemwin in byoffice 
                            if not isemwin or fn[6:8] in myfilters.state_abbrev)
        else:
            bystate = byoffice
        filterednames = (fn for fn, isemwin in bystate)
        return [FileEntry(file_name=osp.join(self.directory, fn)) for fn in filterednames]
    # TODO: how to handle county filters!?!?!    

    def _edit_filters(self, *args, **kwds): 
        self.filters.edit_traits(*args, **kwds)

class LritScanReadFiles(ScanReadFiles):
    directory = CStr(label='Directory', editor=DirectoryEditor(invalid='directory_invalid'),
                     section="Scan and Read Files", config_name="LRIT Directory")

    def _directory_default(self):
        if osp.isdir(config.LRIT_DIRECTORY): return config.LRIT_DIRECTORY
        return super(LritScanReadFiles, self)._directory_default()


class EmwinScanReadFiles(ScanReadFiles):
    directory = CStr(label='Directory', editor=DirectoryEditor(invalid='directory_invalid'),
                     section="Scan and Read Files", config_name="EMWIN Directory")

    def _directory_default(self):
        if osp.isdir(config.EMWIN_DIRECTORY): return config.EMWIN_DIRECTORY
        return super(EmwinScanReadFiles, self)._directory_default()

class ScanReadFilesHandler(Handler):
    dirwatchdog = Instance(QFSWatcher, ())  # notifies when some dir changes
   
    singleview = View(
                    HGroup(
                        Group(
                            UItem('directory'),
                            UReadonly('filtered_list'),
                            HGroup(
                                Readonly('filtered_counter'),
                                Readonly('file_counter')
                            ),
                            show_border = True,
                        ),
                     ),
                    resizable=True,
                    buttons=[],

                 )
    
    dualview = View(
                HGroup(
                    Group(
                        UItem('first.directory'),
                        UCustom('first.filtered_list'),
                        HGroup(
                            Readonly('first.filtered_counter'),
                            Readonly('first.file_counter')
                        ),
                        show_border = True,
                    ),
                    Group(
                        UItem('second.directory'),
                        UCustom('second.filtered_list'),
                        HGroup(
                            Readonly('second.filtered_counter'),
                            Readonly('second.file_counter')
                        ),
                        show_border = True,
                    ),
                 ),
                resizable=True        
               )
    
    def default_traits_view(self): return 'singleview'
    
    def init_info(self, info):
        """Sets up some trait updates b/w dirwatchdog and context objects"""
        # start from either object, or the explicitly set context
        objdict = info.ui.context.get('object', info.ui.context)
        # if this is a HasTraits object (not a dict), get dict from HasTraits
        if isinstance(objdict, ScanReadFiles): objdict = {'object': objdict}
        elif isinstance(objdict, HasTraits): objdict = objdict.trait_get(objdict.trait_names())

        # set watchdog to monitor 'directory' traits; TODO: make this trait name configurable?
        for obj in objdict.itervalues():
            if 'directory' in obj.trait_names():
                obj.on_trait_change(self.dirwatchdog.update_watcher, 'directory')
                self.on_trait_change(obj.update_filelist, 'dirwatchdog.changed')
                obj.update_filelist(obj.directory)  # trigger re-scan

        # now add labels to groups based on obj.label
        cgrps = info.ui.view.content.content[0].content
        for grp in cgrps:
            if 'directory' == grp.content[0].name:
                grp.label = objdict[grp.content[0].object].label
        return True
        
    def configure_traits(self, *args, **kwds):
        """Selects the view based on context length"""
        try:
            if len(kwds['context']) < 2: kwds.setdefault('view', 'singleview')
            else:                        kwds.setdefault('view', 'dualview')
        except:
            pass
        return super(ScanReadFilesHandler, self).configure_traits(*args, **kwds)

class ScanReadFilesTab(ScanReadFilesHandler):
    """Encapsulates a pair of directory selectors with the monitoring handler
    
    This acts a bit like a ModelView where the handler also has the model & view
    
    """

    first = Instance(ScanReadFiles, transient=True)
    second = Instance(ScanReadFiles, transient=True)
    
    def default_traits_view(self): return 'dualview'
    def trait_context(self): return self.trait_get(['first', 'second'])
    
    def __init__(self):
        super(ScanReadFilesTab, self).__init__()
        product_filters = bct.BasicConfigurationsTab()
        self.first = LritScanReadFiles(label='LRIT Products', handler=self, filters=product_filters)
        self.second = EmwinScanReadFiles(label='EMWIN Products', handler=self, filters=product_filters)

    def init_info(self, info):
        super(ScanReadFilesTab, self).init_info(info)
        
        # add menu items with this handler as controller
        from goesradiomenu import top_menu_bar, TopMenuHandler, LocalActionController

        ProdFiltersAction = Action(name='Product Filters...',
                                   action='_open_product_filters',
                                   accelerator='Ctrl+F',
                                   tooltip='configure product filters')

        topgroup = top_menu_bar.find_group('Top')
        topgroup.insert_after(topgroup.find('File'), 
                              Menu('Top', ProdFiltersAction, 
                                    # adding the controller here doesn't work
                                    # b/c its replaced when the UI makes its menu
                                    controller=LocalActionController(handler=self, info=info),
                                    name='&Edit', id='Edit'))
        
                
# QUICK TEST
if __name__ == "__main__":
    _productfilters = bct.BasicConfigurationsTab()
    panel1 = ScanReadFiles(label='PANEL 1', filters=_productfilters)
    ScanReadFilesHandler().configure_traits(context=dict(object=panel1))
    panel2 = ScanReadFiles(label='PANEL 2', filters=_productfilters)
    ScanReadFilesHandler().configure_traits(context=dict(first=panel1, second=panel2))
    # below doesn't work for some reason
    #panel1.configure_traits(
    #    handler=ScanReadFilesHandler(),
    #    context=dict(first=panel1, second=panel2))

